/**
* Label3D18C_.java
* Created on February 2005 by Manuel Guillermo Forero-Vargas
* School of Biosciences. University of Birmingham
* e-mail: m.g.forerovargas@bham.ac.uk, mgforero@yahoo.es
*
* 3D labelling algorith by using 18-connectivity
* Function: This plugin labels white objects (white=255)
*
* This plugin do not create a new image. The maximum number of labeled regions is limited to * 65535.
*
* This plugin is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 
* as published by the Free Software Foundation.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this plugin; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
import ij.*;
import ij.process.*;
import ij.plugin.filter.*;

public class Label3D18C_ implements PlugInFilter
{
    protected   ImageStack  stack;
    public int setup(String arg, ImagePlus imp)
    {
        if (arg.equals("about"))
        {
            showAbout();
            return DONE;
        }
        if (imp==null)
        {
            IJ.noImage();
            return DONE;
        }
        ImageStatistics stats=imp.getStatistics();
        if (stats.histogram[0]+stats.histogram[255]!=stats.pixelCount)
        {
            IJ.error("8-bit binary image (0 and 255) required.");
            return DONE;
        } 
        stack=imp.getStack();
        return DOES_8G+DOES_8C+NO_CHANGES+STACK_REQUIRED;
    }

    public void run(ImageProcessor ip)
    {
        int     i,j,x,y,z;
        boolean	flLabel;
        int[]	neighbors=new int[9];//Store labels of the previous neighbours
        int	iNumNeighbor;//Number of not null neighbours
        int	iNumRegions=0;//Number of regions
        int[]	labels=new int[1048576];//Table of labels
        int	height=ip.getHeight();
        int	width=ip.getWidth();
        int     nslices=stack.getSize();//Number of images
        int[][][]	pixels=new int[nslices][width+2][height+2];
        ImageProcessor ipstack;

        //Creates a copy of the original stack
        for (z=0;z<nslices;z++)
        {
            ipstack=stack.getProcessor(z+1);
            for(y=0;y<height;y++)
                for(x=0;x<width;x++)
                    pixels[z][x+1][y+1]=ipstack.getPixel(x,y);//Original
        }
        //Initialises table of labels
        for(i=1;i<1048576;i++)
            labels[i]=i;
        //Sweeps the image, looking for each voxel of the object and determines to
        //which region it belongs, i.e. either of the previous neighbours and the
        //lower label is assigned to the voxel.
        //  *--*--*--*
        //  | A| A| A|	A. Previous neighbours
        //  *--*--*--*	C. Voxel to analyse
        //  | A| C| P|	P. Next neighbours
        //  *--*--*--*
        //  | P| P| P|
        //  *--*--*--*
        for (z=0;z<nslices;z++)
        {
            IJ.showStatus("Label3D18C:First step Iteration "+z+"/"+nslices+"...");
            IJ.showProgress(z,nslices);
            if (z==0)//Connectivity of the first slice
            {
                iNumRegions=1;
                for (y=1;y<=height;y++)
                    for(x=1;x<=width;x++)
                        if(pixels[0][x][y]>0)
                        {
                            iNumNeighbor=0;
                            //Inspects previous neighbours
                            if(pixels[0][x-1][y-1]>0)
                                neighbors[iNumNeighbor++]=labels[pixels[0][x-1][y-1]];
                            if(pixels[0][x][y-1]>0)
                                neighbors[iNumNeighbor++]=labels[pixels[0][x][y-1]];
                            if(pixels[0][x+1][y-1]>0)
                                neighbors[iNumNeighbor++]=labels[pixels[0][x+1][y-1]];
                            if(pixels[0][x-1][y]>0)
                                neighbors[iNumNeighbor++]=labels[pixels[0][x-1][y]];
                            //Determines if the voxel belongs to a labeled region or a new region
                            if(iNumNeighbor>0)//If voxel belongs to a labeled region...
                            {
                                if (iNumNeighbor>1)
                                    QuickSort.quicksort(neighbors,iNumNeighbor);//Sorts the 4 previous neighbours in ascending order
                                pixels[0][x][y]=neighbors[0];//Takes the smallest label
                                for(i=1;i<iNumNeighbor;i++)//If there is more than one neighbour..
                                    labels[neighbors[i]]=neighbors[0];
                            }
                            else//If the voxel belongs to a new region..
                                pixels[z][x][y]=iNumRegions<1048574?iNumRegions++:1048575;//The number of labeled regions is limited to 1048575
                        }
            }
            else
            {
                for (y=1;y<=height;y++)
                    for(x=1;x<=width;x++)
                        if(pixels[z][x][y]>0)
                        {
                            iNumNeighbor=0;
                            //Inspects the next neighbours
                            if(pixels[z-1][x][y-1]>0)//Inspects voxel of the higher slice
                                neighbors[iNumNeighbor++]=labels[pixels[z-1][x][y-1]];
                            if(pixels[z-1][x-1][y]>0)//Inspects voxel of the higher slice
                                neighbors[iNumNeighbor++]=labels[pixels[z-1][x-1][y]];
                            if(pixels[z-1][x][y]>0)//Inspects voxel of the higher slice
                                neighbors[iNumNeighbor++]=labels[pixels[z-1][x][y]];
                            if(pixels[z-1][x+1][y]>0)//Inspects voxel of the higher slice
                                neighbors[iNumNeighbor++]=labels[pixels[z-1][x+1][y]];
                            if(pixels[z-1][x][y+1]>0)//Inspects voxel of the higher slice
                                neighbors[iNumNeighbor++]=labels[pixels[z-1][x][y+1]];
                            if(pixels[z][x-1][y-1]>0)
                                neighbors[iNumNeighbor++]=labels[pixels[z][x-1][y-1]];
                            if(pixels[z][x][y-1]>0)
                                neighbors[iNumNeighbor++]=labels[pixels[z][x][y-1]];
                            if(pixels[z][x+1][y-1]>0)
                                neighbors[iNumNeighbor++]=labels[pixels[z][x+1][y-1]];
                            if(pixels[z][x-1][y]>0)
                                neighbors[iNumNeighbor++]=labels[pixels[z][x-1][y]];
                            //Determines if the voxel belongs to a labeled region or a new region
                            if(iNumNeighbor>0)//If voxel belongs to a labeled region...
                            {
                                if (iNumNeighbor>1)
                                    QuickSort.quicksort(neighbors,iNumNeighbor);//Sorts the previous neighbours in ascending order
                                pixels[z][x][y]=neighbors[0];//Takes the smallest label
                                for(i=1;i<iNumNeighbor;i++)//If there is more than one neighbour..
                                    labels[neighbors[i]]=neighbors[0];
                            }
                            else//If the voxel belongs to a new region..
                                pixels[z][x][y]=iNumRegions<1048574?iNumRegions++:1048575;//The number of labeled regions is limited to 1048575
                        }
            }
        }
        if (iNumRegions>1048574)
            IJ.showMessage("Warning Maximum number de labels reach..." +iNumRegions);
        //Normalises table of labels
        for(i=2;i<iNumRegions;i++)
            labels[i]=labels[labels[i]];
        //The indexes of the regions in the list are minimised
        for(i=2;i<iNumRegions;i++)
        {
            IJ.showStatus("Label3D18C:Second step Iteration "+i+"/"+iNumRegions+"...");
            IJ.showProgress (i,iNumRegions);
            flLabel=true;
            for(j=i;j<iNumRegions;j++)//If a label is equal to the its position on the list...
                if(labels[j]==i)
                {
                    flLabel=false;
                    break;
                }
            while(flLabel)
            {
                flLabel=false;
                for(j=i;j<iNumRegions;j++)
                    if(labels[j]>i)
                    {
                        labels[j]--;
                        flLabel=true;
                    }
                for(j=i;j<iNumRegions;j++)
                    if(labels[j]==i)
                    {
                        flLabel=false;
                        break;
                    }
            }
        }
        int max=0;
        for (i=1;i<iNumRegions;i++)
            if (max<labels[i])
                max=labels[i];
            if (max>65535)
                IJ.error("Maximum number de labels reach..." +max);
        //Constructs the resultant stack and labels the regions
        ImageStack lstack=new ImageStack(width,height);
        ShortProcessor sip;
        for (z=0;z<nslices;z++)
        {
            sip=(ShortProcessor)stack.getProcessor(z+1).convertToShort(false);
            for(y=0;y<height;y++)
                for(x=0;x<width;x++)
                    sip.putPixel(x,y,(short)labels[pixels[z][x+1][y+1]]);
            lstack.addSlice("",sip);
        }
        ImagePlus iml=new ImagePlus("Label 3D",lstack);
        iml.show();
    }

    void showAbout()
    {
        IJ.showMessage("About Label3D18C_...",
        "This plug-in filter calculates the 18-labelling of binarized stack images.");
    }
}
