/**
* Minimum_errorCell.java
* Created on August 2005 by Manuel Guillermo Forero-Vargas
* School of Biosciences. University of Birmingham
* e-mail: m.g.forerovargas@bham.ac.uk, mgforero@yahoo.es
*
* Minimum error cell thresholding algorithm
* Function: This plugin binarises the image using, as threshold, a given number
* of standard deviations, defined by the user, higher than the peak of the
* second mode, found by a Gaussian approximation
*
* This plugin do not create a new image.
*
* Part of this algorithm is based in the mixture modelling plugin developed by
* Christopher Mei (christopher.mei@sophia.inria.fr) and Maxime Dauphin
*
* This plugin is free software;you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 
* as published by the Free Software Foundation.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY;without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this plugin;if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.measure.*;
import ij.gui.*;

public class Minimum_errorCell implements PlugInFilter, Measurements
{
    private int[] hist;
    private int threshold;
    ImagePlus imp;
    private int q;
    private double mean=0;
    private double sd;
    private double mult;
    static double alpha=3;

    public int setup(String arg, ImagePlus imp)
    {
	if (arg.equals("about"))
        {
            showAbout();
            return DONE;
        }
        //----------------------------------------------------------------------
        //Creates a dialog box, allowing the user to enter the requested
        //number of standard deviatons.
        GenericDialog gd=new GenericDialog("Minimum_error_thresholding",IJ.getInstance());
        gd.addNumericField("Standard_deviations:",alpha,0,2,"");
        gd.showDialog();
        if (gd.wasCanceled())
            return DONE;
        alpha=(double)gd.getNextNumber();
        this.imp=imp;
        return IJ.setupDialog(imp,DOES_8G+DOES_8C);
    }
   
    public void run(ImageProcessor ip)
    {
        double error=0;
        double errorMin=99999999E200;
        double mean_t=0;
        double sd_t=0;
        double mult_t=0;
        ImageStatistics stats;
        int maxEmbrio=0;//Max of the embryo (Peak of the second mode)
        int minq,maxq;//Limits values of the histogram
        int minValley=0,mode;
       	int i,sum;
        int[] bigHist=new int[261];
        
        hist=ip.getHistogram();
        //--------------------------------------------------------------
        //Gets limits of the histogram
        //Returns the first intensity level that appears in the histogram
        stats=ImageStatistics.getStatistics(ip,MODE+MIN_MAX,null);
       minq=(int)stats.min;
        maxq=(int)stats.max;
        mode=stats.mode;
        //----------------------------------------------------------------------
        //Creates a big histogram
        for (i=minq;i<=maxq;i++)
            bigHist[i+5]=hist[i];
        //----------------------------------------------------------------------
        //Gets the maximum of the second peak (Peak of the object)
        for (q=maxq;q>=minq+5;q--)
            if (q==mode)
            {
                maxEmbrio=mode;
                break;
            }
            else if (bigHist[q]>500&(bigHist[q]>=bigHist[q-1])&
            (bigHist[q]>=bigHist[q-2])&(bigHist[q]>=bigHist[q-3])&
            (bigHist[q]>=bigHist[q-4])&(bigHist[q]>=bigHist[q-5])&
            (bigHist[q]>=bigHist[q+1])&(bigHist[q]>=bigHist[q+2])&
            (bigHist[q]>=bigHist[q+3])&(bigHist[q]>=bigHist[q+4])&
            (bigHist[q]>=bigHist[q+5]))
            {
                maxEmbrio=q-5;
                break;
            }
        //----------------------------------------------------------------------
        //If the image has two main peaks, the minimum between the peaks is found
        if (mode<maxEmbrio&mode<minq+15)
        {
            int min=9999999;
            for (q=maxEmbrio;q>mode;q--)//The minValley is found
                if (min>hist[q])
                {
                    min=hist[q];
                    minValley=q;
                }
        }
        else if (mode-minq>5&mode>10)//The histogram could be unimodal or have three peaks, two of them belonging to the embryo
        {
            int min=9999999;
            for (q=maxEmbrio;q>minq+4;q--)//Note:By doing q>minq+4 it is possible to avoid a little peak belonging to the background
                if (min>hist[q])
                {
                    min=hist[q];
                    minValley=q;
                }
        }
        else//If mode==maxEmbrio it means that the histogram is pseudo-exponential
            minValley=minq+5;
        //--------------------------------------------------------------
        //Evaluation of the pdf of the embryo
        for (q=minq;q<maxq;q++)
        {
            mean=0;
            sd=0;
            for(sum=0,i=q;i<=maxq;i++)
            {
                sum+=hist[i];
                mean+=i*hist[i];
            }
            mean/=sum;
            for(i=q;i<=maxq;i++) 
                sd+=hist[i]*(i-mean)*(i-mean);
            sd/=sum;
            mult=hist[(int)(mean+0.5)];
            //--------------------------------------------------------------
            //Evaluation of the error
            for(error=0,i=minValley;i<=maxq;i++)
                error+=(gamma2(i)-hist[i])*(gamma2(i)-hist[i]);
//            error/=256;
            //--------------------------------------------------------------
            if(error<errorMin)
            {
                errorMin=error;
                threshold=q;
                mean_t=mean;
                sd_t=sd;
                mult_t=mult;//Required to show the final curve if runHistogram is true
            }
        }
        mean=mean_t;
        sd=sd_t;
        mult=mult_t;
        threshold=(int)(mean+alpha*Math.sqrt(sd));
        ip.threshold(threshold);
        imp.show();
        imp.updateAndDraw();
    }

    public final double gamma2(int i)
    {
        return (sd==0)?0:mult*Math.exp(-(i-mean)*(i-mean)/(2*sd));
    }

    void showAbout()
    {
        IJ.showMessage("About Minimum_errorCell...",
        "This plug-in filter calculates the mixtureModeling of a 8-bit images.\n");
    }
}
