/**
* Spot_filter.java
* Created on 26 June 2006, 10:15 by Manuel Guillermo Forero-Vargas
* School of Biosciences. University of Birmingham
* e-mail: m.g.forerovargas@bham.ac.uk; mgforero@yahoo.es 
*
* Spot filter algorithm
* Function: Eliminates small bright spots in an image, replacing it by the mean of the values inside the window or the value given by the user.
*
* This plugin do not create a new image
*
* This plugin is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 
* as published by the Free Software Foundation.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this plugin; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.GenericDialog;

public class Spot_filter implements PlugInFilter
 {
    protected   ImagePlus   image;
    protected static int replacement=-1;
    protected static int threshold=200;
    protected static int winwidth=3;
    protected static double percentage=0.66;

    public int setup(String arg, ImagePlus imp)
    {
        if (arg.equals("about"))
        {
            showAbout();
            return DONE;
        }
        if (imp==null)
        {
            IJ.noImage();
            return DONE;
        }
        image=imp;
        //----------------------------------------------------------------------
        //Creates a dialog box, allowing the user to enter the requested values
        GenericDialog gd=new GenericDialog("Spot filter");
        gd.addNumericField("Window side:",winwidth,0,2,"pixels");
        gd.addNumericField("Threshold:",threshold,0,3,"Grey level");
        gd.addNumericField("Percentage:",percentage,2);
        gd.addNumericField("Replace value:",replacement,3);
        gd.showDialog();
        if (gd.wasCanceled())
            return DONE;
        winwidth=(int)gd.getNextNumber();
        threshold=(int)gd.getNextNumber();
        percentage=(double)gd.getNextNumber();
        replacement=(int)gd.getNextNumber();
        //----------------------------------------------------------------------
        //Asks the user to choose to apply the method to the stack or just to the image
        return IJ.setupDialog(imp,DOES_8G+DOES_8C);
    }

    public void run(ImageProcessor ip)
    {
        byte[]	bytepixels=(byte[])ip.getPixels();
        double  mean=0;
        int a,b,c,i,j,offset;
        int x,y;//Variables used to sweep the subimage
        int	height=ip.getHeight();
        int	width=ip.getWidth();
        int midwidth=winwidth/2;//Half the window size
        int bigheight=height+2*midwidth;
        int bigwidth=width+2*midwidth;
	int[]   pixels=new int[bigheight*bigwidth];

        //----------------------------------------------------------------------
        //Creates a copy of the original image with greater size, where the edges are duplicated by reflection.
        //----------------------------------------------------------------------
        //Copies the pixeles in the centre of the image
        for(i=0;i<height;i++)
        {
            a=(i+midwidth)*bigwidth;
            b=i*width;
            c=a+midwidth;
            for(j=0;j<width;j++)
                pixels[c+j]=bytepixels[b+j]&0xff;
        }
        //----------------------------------------------------------------------
        //Duplicates the pixels on the top and bottom of the image
        for (j=midwidth;j<midwidth+width;j++)
        {
            a=2*midwidth-1;//Duplicates top
            for (i=0;i<midwidth;i++)
                pixels[i*bigwidth+j]=pixels[(a-i)*bigwidth+j];
            a=2*bigheight-2*midwidth-1;//Duplicates bottom
            for (i=midwidth+height;i<bigheight;i++)
                pixels[i*bigwidth+j]=pixels[(a-i)*bigwidth+j];
        }
        //----------------------------------------------------------------------
        //Duplicates the pixels on the left and right of the image
        for (i=0;i<bigheight;i++)
        {
            a=2*midwidth-1;//Duplicates left
            b=i*bigwidth;
            for (j=0;j<midwidth;j++)
                pixels[b+j]=pixels[b+(a-j)];
            a=2*bigwidth-2*midwidth-1;//Duplicates right
            c=b+a;
            for (j=midwidth+width;j<bigwidth;j++)
                pixels[b+j]=pixels[c-j];
        }
	//------------------------------------------------------------------------------
        //Applies spot filter
        for (i=midwidth;i<bigheight-midwidth;i++)
        {
            a=i*bigwidth;
            b=(i-midwidth)*width;
            c=b-midwidth;
            for (j=midwidth;j<bigwidth-midwidth;j++)//Gets the pixels of the mobile window
            {
                if ((pixels[a+j])>threshold)
                {
                    for (mean=0,y=i-midwidth;y<=i+midwidth;y++)//Sweeps pixels inside the subwindow
                    {
                        offset=y*bigwidth;
                        for (x=j-midwidth;x<=j+midwidth;x++)
                            mean+=pixels[offset+x];
                    }
                    mean/=winwidth*winwidth;
                    if (mean<(pixels[a+j])*percentage)
                        bytepixels[c+j]=(replacement==-1)?(byte)mean:(byte)replacement;
                }
            }
        }
        //----------------------------------------------------------------------
        image.show();
        image.updateAndDraw();
    }

    void showAbout()
    {
        IJ.showMessage("About Spot_filter...",
        "This plug-in applied the spot filter.");
    }
}
